import sys
sys.path.append('..')

import torch.nn as nn
import torch.nn.functional as F
from torch import optim
from bijou.learner import Learner
from bijou.data import Dataset, DataLoader, DataBunch
from bijou.metrics import accuracy
from bijou.callbacks import EarlyStopping
from datasets import mnist_data

# 1. ------ 数据
x_train, y_train, x_valid, y_valid = mnist_data()
train_ds, valid_ds = Dataset(x_train, y_train), Dataset(x_valid, y_valid)
bs = 128
train_dl = DataLoader(train_ds, batch_size=bs, shuffle=True)
valid_dl = DataLoader(valid_ds, batch_size=bs, shuffle=True)
data = DataBunch(train_dl, valid_dl)

# 2. ------ 模型和优化器
in_dim = data.train_ds.x.shape[1]
out_dim = y_train.max().item()+1
h_dim = 50
model = nn.Sequential(nn.Linear(in_dim, h_dim), nn.ReLU(), nn.Linear(h_dim, out_dim))
opt = optim.SGD(model.parameters(), lr=1.3)


# 3. ------ learner
loss_func = F.cross_entropy
learner = Learner(model, opt, loss_func, data, metrics=[accuracy], callbacks=EarlyStopping(patience=3))

# 4. ------ fit
learner.fit(500)



import sys
sys.path.append('..')

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import optim
from bijou.learner import Learner
from bijou.data import Dataset, DataLoader, DataBunch
from bijou.metrics import accuracy
from bijou.callbacks import Interpreter
from datasets import mnist_data
import matplotlib.pyplot as plt
import numpy as np

if torch.cuda.is_available():
    torch.cuda.manual_seed_all(1)
else:
    torch.manual_seed(1)

# 1. ------ 数据
x_train, y_train, x_valid, y_valid = mnist_data()
x_test = x_valid[:500]
y_test = y_valid[:500]

train_ds, valid_ds, test_ds = Dataset(x_train, y_train), Dataset(x_valid, y_valid), Dataset(x_test, y_test)
bs = 128
train_dl = DataLoader(train_ds, batch_size=bs, shuffle=True)
valid_dl = DataLoader(valid_ds, batch_size=bs, shuffle=True)
test_dl = DataLoader(test_ds, batch_size=bs, shuffle=True)
data = DataBunch(train_dl, valid_dl)

# 2. ------ 模型和优化器
in_dim = data.train_ds.x.shape[1]
out_dim = y_train.max().item()+1
h_dim = 50
model = nn.Sequential(nn.Linear(in_dim, h_dim), nn.ReLU(), nn.Linear(h_dim, out_dim))
opt = optim.SGD(model.parameters(), lr=0.35)


# 3. ------ learner
loss_func = F.cross_entropy
learner = Learner(model, opt, loss_func, data, metrics=[accuracy], callbacks=Interpreter())

# 4. ------ fit
learner.fit(1)

# 5. ------ test
learner.test(test_dl)


def loss(pred, target):
    return F.cross_entropy(pred, target, reduction='none')

scores, xs, ys, preds, indecies = learner.interpreter.top_data(loss, k=10, target='train', largest=True)
print(scores)
print(indecies)
# print(xs)
plt.figure(figsize=[12, 6])
for i in range(10):
    plt.subplot(2, 5, i+1)
    plt.imshow(xs[i].view([28, -1]))
    plt.title(f'{ys[i]} --> {np.argmax(preds[i])}')


# m = learner.interpreter.confusion_matrix()
learner.interpreter.plot_confusion(target='train', class_names=range(10))
learner.interpreter.plot_confusion(target='val', class_names=range(10))
learner.interpreter.plot_confusion(target='test', class_names=range(10))

mcfs = learner.interpreter.most_confused()
print([[c[0], len(c[1])]for c in mcfs])

plt.show()



import sys
sys.path.append('..')

import torch.nn as nn
import torch.nn.functional as F
from torch import optim
from bijou.learner import Learner
from bijou.data import Dataset, DataLoader, DataBunch
from bijou.metrics import accuracy
from bijou.callbacks import LayerAnalysisCallback
from datasets import mnist_data
import matplotlib.pyplot as plt

# 1. ------ 数据
x_train, y_train, x_valid, y_valid = mnist_data()
train_ds, valid_ds = Dataset(x_train, y_train), Dataset(x_valid, y_valid)
bs = 128
train_dl = DataLoader(train_ds, batch_size=bs, shuffle=True)
valid_dl = DataLoader(valid_ds, batch_size=bs, shuffle=True)
data = DataBunch(train_dl, valid_dl)

# 2. ------ 模型和优化器
in_dim = data.train_ds.x.shape[1]
out_dim = y_train.max().item()+1
h_dim = 50
model = nn.Sequential(nn.Linear(in_dim, h_dim), nn.ReLU(), nn.Linear(h_dim, out_dim))
opt = optim.SGD(model.parameters(), lr=0.35)


# 3. ------ learner
loss_func = F.cross_entropy

cbs = [LayerAnalysisCallback(forward=False, hist_span=0.01), LayerAnalysisCallback(forward=True, hist_span=10)]
learner = Learner(model, opt, loss_func, data, metrics=[accuracy], callbacks=cbs)

# 4. ------ fit
learner.fit(3)

# 5. ------ test
# learner.test(valid_dl)

# 6. ------ predict
pred = learner.predict(x_valid)
print(pred.size())

# 7. ------ plot
learner.recorder.plot()
plt.show()



import sys
sys.path.append('..')

import torch.nn.functional as F
import torch.nn as nn
from torch import optim
import matplotlib.pyplot as plt
from bijou.learner import Learner
from bijou.data import DataProcess as dp, Dataset, DataBunch
from datasets import mnist_data


# 1. ------ 获取数据
x_train, y_train, x_valid, y_valid = mnist_data()
train_ds, valid_ds = Dataset(x_train, y_train), Dataset(x_valid, y_valid)
bs = 512
data = DataBunch(*dp.get_dls(train_ds, valid_ds, bs))

# 2. ------ 模型和优化器
in_dim = data.train_ds.x.shape[1]
out_dim = y_train.max().item()+1
h_dim = 50
model = nn.Sequential(nn.Linear(in_dim, h_dim), nn.ReLU(), nn.Linear(h_dim, out_dim))
opt = optim.SGD(model.parameters(), lr=0.5)


# 3. ------ learner
loss_func = F.cross_entropy
learner = Learner(model, opt, loss_func, data)

# 4. ------ find lr
learner.find_lr(max_iter=100)

# 5. ------ plot
plt.show()



import torch
import pickle
import gzip
from pathlib import Path


def mnist_data():
    path = Path(__file__).parent/'example_datasets'/'mnist.pkl.gz'
    with gzip.open(path, 'rb') as f:
        ((x_train, y_train), (x_valid, y_valid), _) = pickle.load(f, encoding='latin-1')
    return map(torch.tensor, (x_train, y_train, x_valid, y_valid))



import sys
sys.path.append('..')

from functools import partial
import torch.nn.functional as F
import torch.nn as nn
from torch.nn import init
from torch import optim
import matplotlib.pyplot as plt
import bijou.callbacks as cbks
import bijou.hook as hook
from bijou.learner import Learner
from bijou.metrics import accuracy
from bijou.modules import Lambda
from bijou.data import DataProcess as dp, Dataset, DataBunch
from datasets import mnist_data


def conv2d(ni, nf, ks=3, stride=2):
    return nn.Sequential(
        nn.Conv2d(ni, nf, ks, padding=ks//2, stride=stride), nn.ReLU())


def get_cnn_layers(out_dim, nfs):
    nfs = [1] + nfs
    return [
        conv2d(nfs[i], nfs[i+1], 5 if i == 0 else 3)
        for i in range(len(nfs)-1)
    ] + [nn.AdaptiveAvgPool2d(1), Lambda(dp.flatten), nn.Linear(nfs[-1], out_dim)]


def get_cnn_model(out_dim, nfs):
    return nn.Sequential(*get_cnn_layers(out_dim, nfs))


def view_tfm(*size):
    def _inner(x):
        return x.view(*((-1,)+size))
    return _inner


if __name__ == '__main__':
    # ------ 获取数据
    x_train, y_train, x_valid, y_valid = mnist_data()
    x_train, x_valid = dp.normalize_to(x_train, x_valid)
    train_ds, valid_ds = Dataset(x_train, y_train), Dataset(x_valid, y_valid)
    bs = 512
    loss_func = F.cross_entropy
    data = DataBunch(*dp.get_dls(train_ds, valid_ds, bs))


    # ------ 模型、优化器
    nfs = [8, 16, 32, 32]
    out_dim = y_train.max().item()+1
    model = get_cnn_model(out_dim, nfs)

    for l in model:  # 模型参数初始化
        if isinstance(l, nn.Sequential):
            init.kaiming_normal_(l[0].weight)
            l[0].bias.data.zero_()

    opt = optim.SGD(model.parameters(), lr=0.4)


    # ------ learner
    cbfs = [partial(cbks.AvgStatsCallback, accuracy)]
    mnist_view = view_tfm(1, 28, 28)
    cbfs.append(partial(cbks.BatchTransformXCallback, mnist_view))
    learner = Learner(model, opt, loss_func, data, callbacks=cbfs)


    # ------ hooks
    def append_stats(hook_, mod, inp, outp):   # pylint: disable=unused-argument
        if not hasattr(hook_, 'stats'):
            hook_.stats = ([], [])
        means, stds = hook_.stats
        means.append(outp.data.mean())
        stds .append(outp.data.std())


    # ------ fit
    with hook.Hooks(model, append_stats) as hooks:
        learner.fit(2)


    # ------ plot
    fig, (ax0, ax1) = plt.subplots(1, 2, figsize=(10, 4))
    for h in hooks:
        ms, ss = h.stats
        ax0.plot(ms[:10])
        ax1.plot(ss[:10])
    plt.legend(range(6))

    fig, (ax0, ax1) = plt.subplots(1, 2, figsize=(10, 4))
    for h in hooks:
        ms, ss = h.stats
        ax0.plot(ms)
        ax1.plot(ss)
    plt.legend(range(6))

    plt.show()



import sys
sys.path.append('..')

from bijou.metrics import accuracy
from bijou.learner import Learner
from bijou.data import DataProcess as dp, Dataset, DataBunch
import torch.nn as nn
import torch.nn.functional as F
from torch import optim
from functools import partial
from datasets import mnist_data
import matplotlib.pyplot as plt


def get_model(in_dim, out_dim, lr=0.5, h_dim=50):
    model = nn.Sequential(nn.Linear(in_dim, h_dim), nn.ReLU(), nn.Linear(h_dim, out_dim))
    return model, optim.SGD(model.parameters(), lr=lr)


def get_model_func(lr=0.5):
    return partial(get_model, lr=lr)


# ------ 获取数据
x_train, y_train, x_valid, y_valid = mnist_data()
train_ds, valid_ds = Dataset(x_train, y_train), Dataset(x_valid, y_valid)
bs = 128
loss_func = F.cross_entropy
data = DataBunch(*dp.get_dls(train_ds, valid_ds, bs))

# ------模型、优化器
in_dim = data.train_ds.x.shape[1]
out_dim = y_train.max().item()+1
model, opt = get_model(in_dim, out_dim)

# ------ learner
learner = Learner(model, opt, loss_func, data, metrics=[accuracy])

# ------ fit
learner.fit_one_cycle(5, high_lr=0.35)

# ------ plot
learner.recorder.plot()
learner.recorder.plot_metrics()
plt.show()



"""
使用了改进的ReLU
"""

import sys
sys.path.append('..')

from functools import partial
import torch
import torch.nn.functional as F
import torch.nn as nn
from torch.nn import init
from torch import optim
import matplotlib.pyplot as plt
import bijou.callbacks as cbks
import bijou.hook as hook
from bijou.learner import Learner
from bijou.metrics import accuracy
from bijou.modules import Lambda, GeneralRelu
from bijou.data import DataProcess as dp, Dataset, DataBunch
from datasets import mnist_data


def conv2d(ni, nf, ks=3, stride=2):
    return nn.Sequential(
        nn.Conv2d(ni, nf, ks, padding=ks//2, stride=stride), nn.ReLU())


def get_cnn_layers(out_dim, nfs, layer, **kwargs):
    nfs = [1] + nfs
    return [layer(nfs[i], nfs[i+1], 5 if i == 0 else 3, **kwargs)
            for i in range(len(nfs)-1)] + [
        nn.AdaptiveAvgPool2d(1), Lambda(dp.flatten), nn.Linear(nfs[-1], out_dim)]


def conv_layer(ni, nf, ks=3, stride=2, **kwargs):
    return nn.Sequential(
        nn.Conv2d(ni, nf, ks, padding=ks//2, stride=stride), GeneralRelu(**kwargs))


def init_cnn(m, uniform=False):
    f = init.kaiming_uniform_ if uniform else init.kaiming_normal_
    for l in m:
        if isinstance(l, nn.Sequential):
            f(l[0].weight, a=0.1)
            l[0].bias.data.zero_()


def get_cnn_model(out_dim, nfs, layer, **kwargs):
    return nn.Sequential(*get_cnn_layers(out_dim, nfs, layer, **kwargs))


def view_tfm(*size):
    def _inner(x):
        return x.view(*((-1,)+size))
    return _inner


if __name__ == '__main__':
    # ------ 获取数据
    x_train, y_train, x_valid, y_valid = mnist_data()
    x_train, x_valid = dp.normalize_to(x_train, x_valid)
    train_ds, valid_ds = Dataset(x_train, y_train), Dataset(x_valid, y_valid)
    bs = 512
    loss_func = F.cross_entropy
    data = DataBunch(*dp.get_dls(train_ds, valid_ds, bs))


    # ------ 模型、优化器
    nfs = [8, 16, 32, 32]
    out_dim = y_train.max().item()+1
    model = get_cnn_model(out_dim, nfs, conv_layer, leak=0.1, sub=0.4, maxv=6.)
    init_cnn(model)

    opt = optim.SGD(model.parameters(), lr=0.4)


    # ------ learner
    cbfs = [partial(cbks.AvgStatsCallback, accuracy)]
    mnist_view = view_tfm(1, 28, 28)
    cbfs.append(partial(cbks.BatchTransformXCallback, mnist_view))

    sched = cbks.combine_scheds([0.2, 0.8], [cbks.sched_cos(0.2, 1.), cbks.sched_cos(1., 0.1)])  # pylint: disable=no-value-for-parameter
    cbfs = cbfs+[partial(cbks.ParamScheduler, 'lr', sched)]

    learner = Learner(model, opt, loss_func, data, callbacks=cbfs)


    # ------ hooks
    def append_stats(hook, mod, inp, outp):  # pylint: disable=unused-argument
        if not hasattr(hook, 'stats'):
            hook.stats = ([], [], [])
        means, stds, hists = hook.stats
        means.append(outp.data.mean().cpu())
        stds .append(outp.data.std().cpu())
        hists.append(outp.data.cpu().histc(40, -7, 7))


    # ------ fit
    with hook.Hooks(model, append_stats) as hooks:
        learner.fit(3)


    # ------ plot
    learner.recorder.plot_lr()
    # learner.recorder.plot_loss()

    fig, (ax0, ax1) = plt.subplots(1, 2, figsize=(10, 4))
    for h in hooks:
        ms, ss, hi = h.stats
        ax0.plot(ms[:10])
        ax1.plot(ss[:10])
        h.remove()
    plt.legend(range(5))

    fig, (ax0, ax1) = plt.subplots(1, 2, figsize=(10, 4))
    for h in hooks:
        ms, ss, hi = h.stats
        ax0.plot(ms)
        ax1.plot(ss)
    plt.legend(range(5))

    def get_hist(h):
        return torch.stack(h.stats[2]).t().float().log1p()
    fig, axes = plt.subplots(2, 2, figsize=(15, 6))
    for ax, h in zip(axes.flatten(), hooks[:4]):
        ax.imshow(get_hist(h), origin='lower')
        ax.axis('off')
    plt.tight_layout()

    def get_min(h):
        h1 = torch.stack(h.stats[2]).t().float()
        return h1[19:22].sum(0)/h1.sum(0)

    fig, axes = plt.subplots(2, 2, figsize=(15, 6))
    for ax, h in zip(axes.flatten(), hooks[:4]):
        ax.plot(get_min(h))
        ax.set_ylim(0, 1)
    plt.tight_layout()

    plt.show()



import sys
sys.path.append('..')

from functools import partial
import torch.nn.functional as F
import torch.nn as nn
from torch import optim
import matplotlib.pyplot as plt
import bijou.callbacks as cbks
from bijou.learner import Learner
from bijou.metrics import accuracy
from bijou.modules import Lambda
from bijou.data import DataProcess as dp, Dataset, DataBunch
from datasets import mnist_data


class SequentialModel(nn.Module):  # pylint: disable=abstract-method
    def __init__(self, *layers):
        super().__init__()
        self.layers = nn.ModuleList(layers)
        self.act_means = [[] for _ in layers]
        self.act_stds = [[] for _ in layers]

    def __call__(self, x):
        for i, l in enumerate(self.layers):
            x = l(x)
            self.act_means[i].append(x.data.mean())
            self.act_stds[i].append(x.data.std())
        return x

    def __iter__(self):
        return iter(self.layers)


def conv2d(ni, nf, ks=3, stride=2):
    return nn.Sequential(
        nn.Conv2d(ni, nf, ks, padding=ks//2, stride=stride), nn.ReLU())


def get_cnn_layers(out_dim, nfs):
    nfs = [1] + nfs
    return [
        conv2d(nfs[i], nfs[i+1], 5 if i == 0 else 3)
        for i in range(len(nfs)-1)
    ] + [nn.AdaptiveAvgPool2d(1), Lambda(dp.flatten), nn.Linear(nfs[-1], out_dim)]


def view_tfm(*size):
    def _inner(x):
        return x.view(*((-1,)+size))
    return _inner


# ------ 获取数据
x_train, y_train, x_valid, y_valid = mnist_data()
x_train, x_valid = dp.normalize_to(x_train, x_valid)
train_ds, valid_ds = Dataset(x_train, y_train), Dataset(x_valid, y_valid)
bs = 512
loss_func = F.cross_entropy
data = DataBunch(*dp.get_dls(train_ds, valid_ds, bs))

# ------ 模型、优化器
nfs = [8, 16, 32, 32]
out_dim = y_train.max().item()+1
model = SequentialModel(*get_cnn_layers(out_dim, nfs))
opt = optim.SGD(model.parameters(), lr=0.9)

# ------ learner
cbfs = [partial(cbks.AvgStatsCallback, accuracy)]
mnist_view = view_tfm(1, 28, 28)
cbfs.append(partial(cbks.BatchTransformXCallback, mnist_view))
learner = Learner(model, opt, loss_func, data, callbacks=cbfs)

# ------ fit
learner.fit(2)

# ------ plot
for l in model.act_means:
    plt.plot(l)
plt.legend(range(6))
plt.show()



import sys
sys.path.append('..')

from functools import partial
import torch
import torch.nn.functional as F
import torch.nn as nn
from torch.nn import init
from torch import optim
import matplotlib.pyplot as plt
import bijou.callbacks as cbks
import bijou.hook as hook
from bijou.learner import Learner
from bijou.metrics import accuracy
from bijou.modules import Lambda
from bijou.data import DataProcess as dp, Dataset, DataBunch
from datasets import mnist_data


def conv2d(ni, nf, ks=3, stride=2):
    return nn.Sequential(
        nn.Conv2d(ni, nf, ks, padding=ks//2, stride=stride), nn.ReLU())


def get_cnn_layers(out_dim, nfs):
    nfs = [1] + nfs
    return [
        conv2d(nfs[i], nfs[i+1], 5 if i == 0 else 3)
        for i in range(len(nfs)-1)
    ] + [nn.AdaptiveAvgPool2d(1), Lambda(dp.flatten), nn.Linear(nfs[-1], out_dim)]


def get_cnn_model(out_dim, nfs):
    return nn.Sequential(*get_cnn_layers(out_dim, nfs))


def view_tfm(*size):
    def _inner(x):
        return x.view(*((-1,)+size))
    return _inner


if __name__ == '__main__':
    # ------ 获取数据
    x_train, y_train, x_valid, y_valid = mnist_data()
    x_train, x_valid = dp.normalize_to(x_train, x_valid)
    train_ds, valid_ds = Dataset(x_train, y_train), Dataset(x_valid, y_valid)
    bs = 512
    loss_func = F.cross_entropy
    data = DataBunch(*dp.get_dls(train_ds, valid_ds, bs))


    # ------ 模型、优化器
    nfs = [8, 16, 32, 32]
    out_dim = y_train.max().item()+1
    model = get_cnn_model(out_dim, nfs)
    for l in model: # 模型参数初始化
        if isinstance(l, nn.Sequential):
            init.kaiming_normal_(l[0].weight)
            l[0].bias.data.zero_()

    opt = optim.SGD(model.parameters(), lr=0.4)


    # ------ learner
    mnist_view = view_tfm(1, 28, 28)
    cbfs = [partial(cbks.BatchTransformXCallback, mnist_view)]

    def accuracy1(*p):
        return accuracy(*p)
    learner = Learner(model, opt, loss_func, data, metrics=[accuracy, accuracy1], callbacks=cbfs)


    #  ------ hooks
    def append_stats(hook_, mod, inp, outp):   # pylint: disable=unused-argument
        if not hasattr(hook_, 'stats'):
            hook_.stats = ([], [], [])
        means, stds, hists = hook_.stats
        means.append(outp.data.mean().cpu())
        stds .append(outp.data.std().cpu())
        hists.append(outp.data.cpu().histc(40, 0, 10))  # histc isn't implemented on the GPU


    # ------ fit
    with hook.Hooks(model, append_stats) as hooks:
        learner.fit(3)


    # ------ plot
    def get_hist(h):
        return torch.stack(h.stats[2]).t().float().log1p()

    fig, axes = plt.subplots(2, 2, figsize=(15, 6))
    for ax, h in zip(axes.flatten(), hooks[:4]):
        ax.imshow(get_hist(h), origin='lower')
        ax.axis('off')
    plt.tight_layout()

    def get_min(h):
        h1 = torch.stack(h.stats[2]).t().float()
        return h1[:2].sum(0)/h1.sum(0)

    fig, axes = plt.subplots(2, 2, figsize=(15, 6))
    for ax, h in zip(axes.flatten(), hooks[:4]):
        ax.plot(get_min(h))
        ax.set_ylim(0, 1)
    plt.tight_layout()

    learner.recorder.plot_metrics()
    plt.show()
